/*******************************************************************/
/*******************************************************************/
/**                                                               **/
/**                        PARI CALCULATOR                        **/
/**                                                               **/
/**                      copyright Babe Cool                      **/
/**                                                               **/
/*******************************************************************/
/*******************************************************************/
/* $Id: gp.c,v 1.914.7.1 1997/01/19 16:18:09 karim Exp karim $ */
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include "genpari.h"
#include "gp.h"

void pari_addfunctions(module **modlist_p, entree *func, char **help);
long marklist(void);

#ifdef READLINE
BEGINEXTERN
#  include <readline.h>
#  ifndef READLINE2
#    include <history.h>
#  endif
   void init_readline (int under_emacs);
   extern int isatty(int);
ENDEXTERN
#else
   void init_readline(int i) { return; }
#endif

void   aide(char *s);
void   commands(int n);
void   err_recover(long numerr);
void   escape(char *tch);
void   free_graph(void);
void   gentypes(void);
void   globs(void);
char** gp_expand_path(char *v);
int    gp_init_entrees(module *modlist, entree **hash, int force);
void   init_defaults(int force);
void   init_graph(void);
void   setdefault(char *a, char *b, int status);
void   setdtimer(char *v, int silent);
void   user_fun(void);
void   write0(long d,char *s);

extern entree *findentry(long n, char *beg, char *end, entree **table);
extern long  avloc,try_to_recover;

#define STACKSIZE    5000  /* history size */
#define MAX_PATH_LEN 1024

static GEN *g;
static char path[MAX_PATH_LEN],*bufferlist[64];
static char prompt[128], thestring[256];
static long prettyp, under_emacs, gpsilent, simplifyflag;
static long chrono, pariecho, primelimit, parisize;
static long bufferindex, tglobal;
static gp_format fmt;

void
usage(char *s)
{
  pariputs("   ### Usage: "); pariputs(s);
  pariputs(" [-f] [-s stacksize] [-p primelimit] [-b buffersize]\n");
  exit(0);
}

/* must be called BEFORE init() */
void
gp_preinit(int force)
{
  static char *dflt_prompt;

  if (force)
  {
    parisize = 4000000; primelimit = 500000;
    dflt_prompt = "? ";
#ifdef macintosh
# ifndef __MWERKS__
    dflt_prompt = "?\n";
    parisize = 1000000; primelimit = 200000;
# endif
#else
# ifdef LONG_IS_64BIT 
    parisize = 10000000;
# endif
#endif
  }
  strcpy(prompt, dflt_prompt);
#ifdef UNIX
  strcpy(path,".:~:~/gp");
#else
  strcpy(path,".");
#endif
  prettyp = PRETTYMAT;
  tglobal = bufferindex = 0;
  simplifyflag = 1;
  under_emacs = chrono = pariecho = 0; 
  fmt.format = 'g'; fmt.field = 0;
#ifdef LONG_IS_64BIT
  fmt.nb = 38;
#else
  fmt.nb = 28;
#endif
  g = (GEN *) xmalloc(sizeof(GEN)*STACKSIZE);
}

#define GET_SEP_SIZE 128
#define separe(c)  ( (c)==';' || (c)==':' )
/* Return all chars, up to next separator */
char*
get_sep(char *t)
{
  static char s[GET_SEP_SIZE];
  int i=0, outer=1, I=GET_SEP_SIZE-1;
  char c;
  
  while (i<I)
  {
    c = s[i++] = *t++;
    if (c == '"' && (outer || s[i-2] != '\\')) outer = !outer;
    if (!*t || (outer && separe(*t)) ) break;
  }
 /* if (i==GET_SEP_SIZE) ... */
  s[i]=0; return s;
}

void
err_initrc(char *s)
{
  pariputs("   ### in .gprc : ");
  pariputs(s); pariputc('\n');
  exit(1);
}

#ifdef UNIX
#  define _INCLUDE_POSIX_SOURCE
#  include <pwd.h>
#  include <unistd.h>
#  include <sys/types.h>

#define BIGENOUGH 8192
char *
gp_initrc()
{
  char  str[BIGENOUGH], *precmd, *v, *lim, *home = getenv("HOME");
  FILE *file;

  if (! home)
  {
    struct passwd *p = getpwuid(geteuid());
    if (! p) return NULL;
    home = p->pw_dir;
  }
  sprintf(str,"%s/.gprc",home);
  if (! (file=fopen(str,"r")) ) return NULL;

  precmd = (char *) malloc(BIGENOUGH);
  v = precmd; lim = v + BIGENOUGH;
  *precmd=0; errfile=stderr; try_to_recover=0;
  for(;;)
  {
    long len;
    char *s1,*s2, *s=str;

    if (!fgets(s,BIGENOUGH,file)) break;
    if (strlen(s) >= BIGENOUGH-1)
      err_initrc("Try inserting some <RET> in the file.");
    do
    {
      filtre(s); s1=get_sep(s); len=strlen(s1); 
      if (!len) continue;
      if (len > GET_SEP_SIZE) err_initrc("Line too long.");
      s+=len; 
      if (!strncmp(s1,"read\"",5))
      {
	char *p = s1+4;

	sprintf(v,"read(\""); v+=6;
	v = readstring(&p,v,lim,0);
	sprintf(v,"\");"); v+=3;
	continue;
      }
      s2=s1; while (*s2 && *s2 != '=') s2++;
      if (*s2 != '=') err_initrc("Missing '='.");
      *s2++ = 0; setdefault(s1,s2,INITRC);
    } while (*s++);
  }
  fclose(file); try_to_recover=1;
  return precmd;
}
#else
  char *gp_initrc() { return NULL; }
#endif
#undef BIGENOUGH

long
get_int(char *s, long dflt)
{
  char *p=get_sep(s);
  long n=atol(p);

  if (*p == '-') p++;
  while(isdigit(*p)) { p++; dflt=n; }
  if (*p) 
    err(talker2,"I was expecting an integer here: ", s, s);
  return dflt;
}

int
init_brace(char *tch)
{
  int wait_for_brace = 0;
  char *s=tch;

  while (isspace(*tch)) tch++;
  if (*tch == '{') {*tch=' '; wait_for_brace = 1;}
  do_filtre(s);
  return wait_for_brace;
}

/* status = NOPRINT   don't print.
 * status = REGULAR   print elapsed time. (usual chrono = 1)
 * status = LAST      print last elapsed time. (##)
 * status = INTERRUPT received a SIGINT.
 */
char *
do_time(long status)
{
  static long last = 0;
  long delay = timer();

  switch(status)
  {
    case NOPRINT:
      last=delay; return NULL;
    case REGULAR:
      sprintf(thestring,"time = "); break;
    case LAST:
      sprintf(thestring,"  ***   last result computed in ");
      delay=last; break;
    case INTERRUPT:
      sprintf(thestring,"user interrupt after "); break;
  }
  last = delay;

  if (delay >= 3600000)
  {
    sprintf(thestring + strlen(thestring), "%ldh, ", delay / 3600000);
    delay %= 3600000;
  }
  if (delay >= 60000)
  {
    sprintf(thestring + strlen(thestring), "%ldmn, ", delay / 60000);
    delay %= 60000;
  }
  if (delay >= 1000)
  {
    sprintf(thestring + strlen(thestring), "%ld,", delay / 1000);
    delay %= 1000;
    if (delay < 100)
      sprintf(thestring + strlen(thestring), "%s", (delay<10)? "00": "0");
  }
  sprintf(thestring + strlen(thestring), "%ld ms", delay);
  return thestring;
}

void
do_echo(char *buf)
{
  if (pariecho)
  {
    pariputs(buf);
    pariputc('\n');
  }
  if (logfile)
  {
    fputs(buf, logfile);
    fputc('\n',logfile);
  }
}

void
fix_buffer(char **buf, char **s, long *ptr_old, long toadd)
{
  long len = *s - *buf, newlen = len+toadd;
  char *gpbuffer = *buf;

  if (newlen > paribuffsize)
  {
    gpbuffer = (char *) xrealloc(gpbuffer,newlen,*ptr_old);
    *ptr_old = paribuffsize = newlen;
    *s = gpbuffer+len; *buf = gpbuffer;
    bufferlist[bufferindex] = gpbuffer;
  }
}

void
gp_output(GEN z)
{
  long tz=typ(z);

  if (is_matvec_t(tz) && prettyp == PRETTY) pariputc('\n');
  if (fmt.nb >= 0 && is_intreal_t(tz))
    ecrire(z, fmt.format, fmt.nb, fmt.field);
  else
    switch(prettyp)
    {
      case RAW       : brute(z, fmt.format, fmt.nb); break;
      case PRETTYMAT : matbrute(z, fmt.format, fmt.nb); break;
      case PRETTY    : sor(z, fmt.format, fmt.nb, fmt.field); break;
      case TEX       : texe(z, fmt.format, fmt.nb); break;
    }
}

void
gp_quit()
{
  freeall(); free_graph();
  while (bufferindex) free((void *)bufferlist[--bufferindex]);
  if (INIT_SIG)
  {
    signal(SIGINT,SIG_DFL);
    signal(SIGSEGV,SIG_DFL);
#ifdef SIGBUS
    signal(SIGBUS,SIG_DFL);
#endif
  }
  pariputs("Good bye!\n"); exit(0);
}

void
gp_sighandler(int sig)
{
  char *msg;
  switch(sig)
  {
    case SIGINT:
      if (infile != stdin)
      {
	int file_depth = switchin(NULL);
	if (file_depth == -1) { freeall(); exit(0); }
      }
      msg = do_time(INTERRUPT);
      break;

    case SIGSEGV:
      msg="segmentation fault: bug in GP (please report)";
      break;

#ifdef SIGBUS
    case SIGBUS:
      msg="bus error: bug in GP (please report)";
      break;
#endif
  }
  signal(sig,gp_sighandler);
  err(talker,msg);
}

/*
 * GP main loop.
 *  status != MAIN: we are doing an immediate read (with \r, read, etc.)
 *  status == SILENT: keep silent.
 */
void
gp_main_loop(long status)
{
  static long tloc,listloc;
  long av, tets, len, read_done, oldsize=paribuffsize;
  char  *gpbuffer, *tch;
  GEN  z;

  if (bufferindex > 63) err(talker,"Too many nested files");
  bufferlist[bufferindex++] = gpbuffer = (char *) xmalloc(paribuffsize);
  for(;;)
  {
    int done_already, wait_for_brace;
#ifdef UNIX /* This is probably standard Ansi C. */
    char promptbuf[80];
    time_t t=time(NULL);

    strftime(promptbuf,79,prompt,localtime(&t));
#else
#  define promptbuf prompt
#endif
   /* setdefault can't change the buffer on the fly ==> check every loop. */
    fix_buffer(&gpbuffer,&gpbuffer,&oldsize,paribuffsize);
    avloc = avma; tloc = tglobal; listloc = marklist();
    if (setjmp(environnement))
    {
      avma = avloc; tglobal = tloc;
      try_to_recover=0; recover(listloc); try_to_recover=1;
    }

#ifdef READLINE
#define BIGENOUGH 1024
    if (isatty(fileno(infile)) || (under_emacs && !bufferindex))
    {
      static char tmp_history[BIGENOUGH];
      char *rlbuffer, *index = tmp_history, *lim = index+BIGENOUGH-1;

      done_already=1;
      if (logfile) fputs(promptbuf, logfile);
      for(;;)
      {
	rlbuffer = readline(promptbuf);
	if (!rlbuffer)
	{
	  /* EOF */
	  pariputs("\n"); strcpy(gpbuffer,"\\q");
	  break;
	}
	if (*rlbuffer) break;
      } 

      if (rlbuffer)
      {
	len = strlen(rlbuffer);
	if (index+len>=lim)
	  err(talker,"you should write this to a file instead");
	strcpy(index,rlbuffer); index+=len;
	wait_for_brace = init_brace(rlbuffer);
	fix_buffer(&gpbuffer, &gpbuffer, &oldsize, strlen(rlbuffer)+1);
	strcpy(gpbuffer,rlbuffer); free(rlbuffer);
      }

      /* gpbuffer is never empty at this point */
      tch = gpbuffer;
      if (*tch != '?')
	for(;;)
	{ 
	  tch = gpbuffer + strlen(gpbuffer) - 1;
	  if (wait_for_brace && *tch == '}') {*tch=0; break;}

	  if (*tch != '\\')
	  {
	    /* Recall: comments have already been suppressed */
	    if (!wait_for_brace) break;
	    tch++;
	  } else *tch=0;

	  /* read continuation line */
	  rlbuffer = readline(""); if (!rlbuffer) break;
	  len = strlen(rlbuffer);
	  if (index+len>=lim)
	    err(talker,"you should write this to a file instead");
	  strcpy(index,rlbuffer); index+=len;
	  do_filtre(rlbuffer);
	  fix_buffer(&gpbuffer, &tch, &oldsize, strlen(rlbuffer)+1);
	  strcpy(tch,rlbuffer); free(rlbuffer);
	}
      add_history(tmp_history);
# if defined (SOLARIS) && defined (READLINE2)
      sigrelse(SIGINT); /* bug in readline, which does not unblock ^C */
# endif
    }
    else
#undef BIGENOUGH
#endif /* defined(READLINE) */
    {
      done_already = 0;
      if (infile==stdin) pariputs(promptbuf);
      if (!fgets(gpbuffer, paribuffsize, infile))
      {
	int file_depth = switchin(NULL);
	if (file_depth == -1) gp_quit();
	if (status != MAIN) break; else continue;
      }
      read_done = (gpbuffer[strlen(gpbuffer)-1] == '\n');
      wait_for_brace = init_brace(gpbuffer);
      if (! *gpbuffer) /* buffer empty */
      {
	if (! wait_for_brace) continue;
	do
	{
	  fgets(gpbuffer, paribuffsize, infile);
	  do_filtre(gpbuffer);
	} while (! *gpbuffer);
      }
    }
    do_echo(gpbuffer);

    switch(*gpbuffer)
    {
      case '?':
	aide(get_sep(gpbuffer+1)); pariputc('\n'); continue;

      case '\\':
	escape(gpbuffer+1); continue;

      case '#':
        tch = gpbuffer+1;
	if (*tch)
	{
	  int last=0;

	  if (*tch == '#') { tch++; last=1; }
	  if (*tch && !separe(*tch))
	    err(caracer1,tch,gpbuffer);
	  if (last)
	  {
	    pariputs(do_time(LAST));
	    pariputs(".\n"); continue;
	  }
	}
	chrono=1-chrono; setdtimer("",ACKNOWLEDGE); continue;

      default:
        if (done_already) break;
	for(;;)
	{
	  tch = gpbuffer + strlen(gpbuffer) - 1;

	  if (wait_for_brace && *tch == '}') {*tch=0; break;}
	  if (*tch != '\\')
	  {
	    if (!wait_for_brace && read_done) break;
	    tch++;
	  } 
	  len = paribuffsize - (tch-gpbuffer);
	  if (len <= 1024)
	  {
	    fix_buffer(&gpbuffer, &tch, &oldsize, paribuffsize);
	    len = paribuffsize - (tch-gpbuffer);
	  }
	  if (!fgets(tch, len, infile)) break;
	  read_done = (gpbuffer[strlen(gpbuffer)-1] == '\n');
	  do_filtre(tch); do_echo(tch);
	}
    }
    tch = gpbuffer + strlen(gpbuffer) - 1;
    gpsilent = (*tch == ';' || *tch == ':' || status == SILENT);
    tch = gpbuffer;
    fflush(outfile); if (logfile) fflush(logfile);

    if (!tglobal) timer2();
    timer(); av=avma; z = readseq(tch);
    if (chrono)
    { 
      pariputs(do_time(REGULAR)); pariputs(".\n");
    }
    else do_time(NOPRINT);
    if (z == gnil) continue;

    if (tglobal>=STACKSIZE)
      err(talker,"GP history overflows, please restart or use \\k");
    len=isonstack(z);
    if (simplifyflag) 
    {
      tets=avma; z=simplify(z);
      if (len) z = gerepile(av,tets,z);
    }
    else if (!len) z=gcopy(z);
    g[0] = g[++tglobal] = z; 
    if (!gpsilent)
    {
      print1("%%%ld = ",tglobal);
      gp_output(z); pariputc('\n'); 
    }
  }
  free(gpbuffer); bufferindex--;
}

static char *
centerstring(char *s)
{
  long i, pad = (75-strlen(s))>>1;
  char *u = thestring;
  
  for(i=0; i<pad; i++) *u++ = ' ';
  while(*s) *u++ = *s++;
  *u++='\n'; *u=0;
  return thestring;
}

static void
print_version()
{
  pariputs(centerstring(PARIVERSION));
  pariputs(centerstring(PARIINFO));
}

/*
 * history management function.
 *   If type = -2, called from freeall()
 *   If type = -1, called from allocatemoremem()
 *   If type = 0 , called from %num in anal.c:truc()
 *   If type > 0 , called from %` in anal.c:truc()
 */
GEN
gp_history(long p, long type, char *old, char *entrypoint)
{
  if (type==-2) { free((void *)g); return NULL; }
  if (type==-1)
  {
    long v, declg = p;
    GEN topold = (GEN) entrypoint, avmaold = (GEN) old;

    for (v=0; v<=tglobal; v++)
      if ( g[v] < topold && g[v] >= avmaold ) g[v] += declg;
    return NULL;
  }

  if (type || ! tglobal) /* called from %` or history empty */
  {
    if (p>=tglobal) err(referer1,old,entrypoint);
    return g[tglobal-p];
  }

  if (p>tglobal) err(referer2,old,entrypoint);
  return g[p];
}

int
main(long argc, char **argv)
{
  int i = 1, gprc = 1;
  char *b = NULL, *p = NULL, *s = NULL, *precmd = NULL;

  init_defaults(1); gp_preinit(1); 
#ifdef __MWERKS__
  argc = ccommand(&argv);
#endif
  outfile=stderr;
  while (i<argc)
  {
    char *tch = argv[i++];

    if (*tch++ != '-') usage(argv[0]);
    switch(*tch)
    {
      case 'b':
	if (i==argc) usage(argv[0]);
	b = argv[i++]; break;
      case 'p':
	if (i==argc) usage(argv[0]);
	p = argv[i++]; break;
      case 's':
	if (i==argc) usage(argv[0]);
	s = argv[i++]; break;

      case 'e': 
	if (!strncmp(tch,"emacs",5)) under_emacs = 1;
        break;
      case 'f':
	gprc = 0; break;
      default:
	usage(argv[0]);
    }
  }
  if (gprc) precmd = gp_initrc();
  if (b) paribuffsize = atoi(b);
  if (p) primelimit = atoi(p);
  if (s) parisize = atoi(s);

  init_readline(under_emacs); outfile=stdout;
  pari_addfunctions(&pari_modules, functions_gp,helpmessages_gp);
  pari_addfunctions(&pari_oldmodules, functions_oldgp,helpmessages_oldgp);

  init_graph();
  INIT_SIG_off;
  pari_init(parisize, primelimit);

#ifdef SIGBUS
  signal(SIGBUS,gp_sighandler);
#endif
  signal(SIGINT,gp_sighandler);
  signal(SIGSEGV,gp_sighandler);

  gp_history_fun = gp_history;
  gp_expand_path(path);

  if (precmd)
  {
    i=chrono; chrono=0; 
    lisseq(precmd); free(precmd);
    chrono=i;
  }
  
  print_version();
  pariputs("\nCopyright 1989-1997 by C.Batut, K.Belabas, D.Bernardi, ");
  pariputs("H.Cohen and M.Olivier.\n\nType ? for help.\n\n");
  globs();

  gp_main_loop(MAIN); gp_quit();
  return 0; /* not reached */
}

/********************************************************************/
/*                                                                  */
/*                       GESTION des DEFAULTS                       */
/*                                                                  */
/********************************************************************/

void
setdrealprecision(char *v, int status)
{
  if (*v)
  {
    fmt.nb = get_int(v, fmt.nb);
    prec = (long) fmt.nb*K1 + 3;
  }
  if (status != ACKNOWLEDGE) return;
  print1("   realprecision = %ld significant digits\n",fmt.nb);
}

void
setdseriesprecision(char *v, int status)
{
  if (*v) precdl = get_int(v, precdl);
  if (status != ACKNOWLEDGE) return;
  print1("   seriesprecision = %ld significant terms\n",precdl);
}

void 
setdformat(char *v, int status)
{
  if (*v)
  {
    char c = *v;

    if (c!='e' && c!='f' && c!='g')
      err(talker2,"bad format: ",v,v);
    fmt.format = c; v++;

    if (isdigit(*v)) 
      { fmt.field=atol(v); while (isdigit(*v)) v++; }
    if (*v++ == '.')
    {
      if (*v == '-') fmt.nb = -1;
      else
	if (isdigit(*v)) fmt.nb=atol(v);
    }
  }
  if (status != ACKNOWLEDGE) return;
  print3("   format = %c%ld.%ld\n", fmt.format, fmt.field, fmt.nb);
}

void
setdprompt(char *v, int status)
{
  if (*v)
  {
    if (status == INITRC) readstring(&v, v, v + GET_SEP_SIZE,1);
    strcpy(prompt,v);
#ifdef macintosh
    strcat(prompt,"\n");
#endif
  }
  if (status != ACKNOWLEDGE) return;
  print1("   prompt = \"%s\"\n",prompt);
}

void
setdpath(char *v, int status)
{
  if (*v)
  {
    if (status == INITRC) readstring(&v, v, v + GET_SEP_SIZE,1);
    if (strlen(v) >= MAX_PATH_LEN) err(talker,"Path too long");
    strcpy(path,v); if (status == INITRC) return;
    gp_expand_path(path);
  }
  if (status != ACKNOWLEDGE) return;
  print1("   path = \"%s\"\n",path);
}

void 
setdoutput(char *v, int status)
{
  if (*v)
  {
    long l=get_int(v,0);
    while (isdigit(*v)) v++;
    if (*v || l>2 || l<0)
      err(talker,"setdefault: expected value 0, 1, or 2");
    prettyp=l;
  }
  if (status != ACKNOWLEDGE) return;
  switch(prettyp)
  {
    case PRETTY    : pariputs("   output = 2 (prettyprint)\n"); break;
    case PRETTYMAT : pariputs("   output = 1 (prettymatrix)\n"); break;
    case RAW       : pariputs("   output = 0 (raw)\n");
  }
}

#define test_flag(l,v) \
  if ((*(v)) || (l)>1 || (l)<0) \
   err(talker,"setdefault: expected value 0 or 1.")

void 
setdtimer(char *v, int status)
{
  if (*v)
  {
    long l=get_int(v,0);
    while (isdigit(*v)) v++;
    test_flag(l,v);
    chrono=l;
  }
  if (status != ACKNOWLEDGE) return;
  pariputs( (chrono)
    ? "   timer = 1 (on)\n"
    : "   timer = 0 (off)\n"
  );
}

void 
setdecho(char *v, int status)
{
  if (*v)
  {
    long l=get_int(v,0);
    while (isdigit(*v)) v++;
    test_flag(l,v);
    pariecho=l;
  }
  if (status != ACKNOWLEDGE) return;
  pariputs( (pariecho)
    ? "   echo = 1 (on)\n"
    : "   echo = 0 (off)\n"
  );
}

void 
setdsimplify(char *v, int status)
{
  if (*v)
  {
    long l=get_int(v,0);
    while (isdigit(*v)) v++;
    test_flag(l,v);
    simplifyflag=l;
  }
  if (status != ACKNOWLEDGE) return;
  pariputs( (simplifyflag)
    ? "   simplify = 1 (automatic simplification on)\n"
    : "   simplify = 0 (automatic simplification off)\n"
  );
}

void
setdlog(char *v, int status)
{
  if (*v)
  {
    long l=get_int(v,0);
    while (isdigit(*v)) v++;
    test_flag(l,v);
    if (l)
    {
      logfile = fopen(current_logfile, "a");
      if (!logfile) err(outloger1);
    } else { fclose(logfile); logfile=NULL; }
  }
  if (status != ACKNOWLEDGE) return;
  pariputs( logfile ? "   log = 1 (logging on)\n"
                    : "   log = 0 (logging off)\n"
  );
}

void
setdlogfile(char *v, int status)
{
  if (*v)
  {
    if (logfile) fclose(logfile);
    if (status == INITRC) readstring(&v, v, v + GET_SEP_SIZE,1);
    v=expand_tilde(v,thestring);
    strncpy(current_logfile,v,128);
    current_logfile[127]=0;
  }
  if (status != ACKNOWLEDGE) return;
  print1("   logfile = \"%s\"\n",current_logfile);
}

void
setdpsfile(char *v, int status)
{
  if (*v)
  {
    if (status == INITRC) readstring(&v, v, v + GET_SEP_SIZE,1);
    v=expand_tilde(v,thestring);
    strncpy(current_psfile,v,128);
    current_logfile[127]=0;
  }
  if (status != ACKNOWLEDGE) return;
  print1("   psfile = \"%s\"\n",current_psfile);
}

void
setddebug(char *v, int status)
{
  if (*v) DEBUGLEVEL=get_int(v,0);
  if (status != ACKNOWLEDGE) return;
  print1("   debug = %ld\n", DEBUGLEVEL);
}

void
setdcompatible(char *v, int status)
{
  if (*v)
  {
    int res;
    compatible=get_int(v,0);
    if (status == INITRC) return;
    if (new_fun_set)
      res = gp_init_entrees(pari_modules, functions_hash, 0);
    else
      res = gp_init_entrees(pari_oldmodules, functions_hash, 0);
    if (res)
      fprintferr("  *** Warning: variables re-initialized\n");
  }
  if (status != ACKNOWLEDGE) return;
  switch(compatible)
  {
    case NONE: 
      pariputs("   compatible = 0 (no backward compatibility)\n");
      break;
    case WARN:
      pariputs("   compatible = 1 (warn when using obsolete functions)\n");
      break;
    case OLDFUN:
      pariputs("   compatible = 2 (use old functions, don't ignore case)\n");
      break;
    case OLDALL:
      pariputs("   compatible = 3 (use old functions, ignore case )\n");
      break;
  }
}

void
setdprimelimit(char *v, int status)
{
  if (*v)
  {
    long tmp=get_int(v,0);
    unsigned char *ptr;
    if (status != INITRC) {ptr=initprimes(tmp); free(diffptr); diffptr=ptr;}
    primelimit=tmp;
  }
  if (status != ACKNOWLEDGE) return;
  print1("   primelimit = %ld\n", primelimit);
}

void
setdparisize(char *v, int status)
{
  if (*v)
  {
    long newsize=get_int(v,0);
    if (status != INITRC)
    {
      allocatemoremem(newsize);
      parisize=newsize;
      if (status == ACKNOWLEDGE)
	print1("   parisize = %ld\n", parisize);
      longjmp(environnement,errpile);
    }
    parisize=newsize;
  }
  if (status != ACKNOWLEDGE) return;
  print1("   parisize = %ld\n", parisize);
}

void
setdbuffersize(char *v, int status)
{
  if (*v) paribuffsize=get_int(v,0);
  if (status != ACKNOWLEDGE) return;
  print1("   buffersize = %ld\n", paribuffsize);
}

default_type gp_default_list[] =
{
  {"buffersize",(void*)setdbuffersize},
  {"compatible",(void*)setdcompatible},
  {"debug",(void*)setddebug},
  {"echo",(void*)setdecho},
  {"format",(void*)setdformat},
  {"log",(void*)setdlog},
  {"logfile",(void*)setdlogfile},
  {"output",(void*)setdoutput},
  {"parisize",(void*)setdparisize},
  {"path",(void*)setdpath},
  {"primelimit",(void*)setdprimelimit},
  {"prompt",(void*)setdprompt},
  {"psfile",(void*)setdpsfile},
  {"realprecision",(void*)setdrealprecision},
  {"seriesprecision",(void*)setdseriesprecision},
  {"simplify",(void*)setdsimplify},
  {"timer",(void *)setdtimer},
  {NULL,NULL} /* sentinel */
};

void
help_default()
{
  long n=0;
  default_type dflt;

  do
  {
    dflt=gp_default_list[n++];
    ((void (*)(ANYARG)) dflt.fun)("",0);
  } while (gp_default_list[n].name);
}

/* silent = 0: print a message
 * silent = 1: don't print a message
 * silent = 2: don't print a message (from gprc)
 */
void
setdefault(char *s,char *v, int status)
{
  long n=0;
  default_type dflt;

  if (!*s) { help_default(); return; }

  do
  {
    dflt=gp_default_list[n++];
    if (!strcmp(s,dflt.name))
    {
      ((void (*)(ANYARG)) dflt.fun)(v,status);
      return;
    }
  } while (gp_default_list[n].name);

  sprintf(thestring,"Unknown default: %s",s);
  err(talker,thestring);
}



/********************************************************************/
/**                                                                **/
/**                    COMMANDES COMMENCANT PAR \                  **/
/**                                                                **/
/**                     ET ANALOGUES DANS ANAL.C                   **/
/**                                                                **/
/********************************************************************/

void
killall()
{
  long i;
  
  free_graph();
  for (i=1;i<=NUMPRTBELT;i++)
    if(!gcmp1((GEN)primetab[i]))
    {
      killbloc((GEN)primetab[i]);
      primetab[i]=un;
    }

  gpi = geuler = bernzone = (GEN)0;
  for(i=0; i<MAXVAR; i++) ordvar[i]=i;
  setlg(polvar,1);
  for(i=1; i<=MAXVAR; i++)
    polvar[i] = evaltyp(t_VEC)+evalpere(MAXPERE)+evallg(1);  

  init_defaults(1); gp_preinit(0); init_graph(); 
  gp_init_entrees(pari_modules, functions_hash, 1);
}  

void
escape(char *tch)
{
  if (compatible != NONE)
  {
    char *s=tch;

    while (*s && *s != '=') s++;
    if (*s)
    {
      int done = 1, len = s-tch;
      
      s = get_sep(s+1);
      if (!strncmp(tch,"precision",len))
	setdrealprecision(s,ACKNOWLEDGE);
      else if (!strncmp(tch,"serieslength",len))
	setdseriesprecision(s,ACKNOWLEDGE);
      else if (!strncmp(tch,"format",len))
	setdformat(s,ACKNOWLEDGE);
      else if (!strncmp(tch,"prompt",len))
	setdprompt(s,ACKNOWLEDGE);
      else done=0;

      if (done) return;
    }
  }

  switch (*tch++)
  {
    case 'a':
      brute(g[get_int(tch, tglobal)], fmt.format, -1);
      pariputc('\n'); break;
    case 'b':
      sor(g[get_int(tch, tglobal)], fmt.format, -1, fmt.field);
      pariputc('\n'); break;
    case 'c': commands(-1); break;
    case 'd': help_default(); break;
    case 'e': 
      pariecho=1-pariecho;
      setdecho(pariecho?"1":"0",ACKNOWLEDGE); break;
    case 'g': setddebug(tch,ACKNOWLEDGE); break;
    case 'k': killall(); globs(); break;
    case 'l': setdlog(logfile?"0":"1",ACKNOWLEDGE); break;
    case 'm':
      matbrute(g[get_int(tch, tglobal)], fmt.format, -1);
      pariputc('\n'); break;
    case 'p': setdrealprecision(tch,ACKNOWLEDGE); break;
    case 'q': gp_quit(); break;
    case 'r': switchin_byname(get_sep(tch)); break;
    case 's': etatpile(get_int(tch, 0)); break;
    case 't': gentypes(); break;
    case 'u': user_fun(); break;
    case 'v': print_version(); break;
    case 'w':
    {
      char *s, *old=tch;
      long d;

      for (d=0;isdigit(*tch);) d = 10*d + *tch++ - '0';
      if (d>tglobal) err(referer2,old,old-2);
      s=get_sep(tch); 
      write0(d,s); break;
    }
    case 'x': voir(g[tglobal], get_int(tch, -1)); break;
    case 'y':
      simplifyflag=1-simplifyflag;
      setdsimplify(simplifyflag?"1":"0",ACKNOWLEDGE); break;
    default: err(caracer1,tch-1,tch-2);
  }
}

/********************************************************************/
/**                                                                **/
/**          AFFICHAGE TYPES, COMMANDES AIDES ET GLOBALES          **/
/**                                                                **/
/********************************************************************/

void
globs()
{
  char s[70];
  
  setdrealprecision("",ACKNOWLEDGE);
  setdseriesprecision("",ACKNOWLEDGE);
  setdformat("",ACKNOWLEDGE);
  sprintf(s,"\nparisize = %ld, primelimit = %ld, buffersize = %ld",
               parisize, primelimit, paribuffsize);
  pariputs(s); pariputc('\n');
}

static int
compare_str(char **s1, char **s2)
{
  return (strcmp (*s1, *s2));
}

#define TERM_WIDTH 79 /* we suppose the terminal width is 80 columns */

/* Print all elements of list in columns, pausing every nbli lines
 * if nbli is non-zero.
 *
 * list is a NULL terminated list of function names
 */
void 
print_fun_list(char **list, int nbli)
{
  int i=0, j=0, maxlen=0, nbcol,len,v;

  while (list[i]) i++;
  qsort (list, i, sizeof(char *), (int(*)(ANYARG)) compare_str);

  for (v=0; list[v]; v++)
  {
    len = strlen(list[v]);
    if (len > maxlen) maxlen=len;
  }
  maxlen+=1; nbcol= TERM_WIDTH / maxlen;
  if (nbcol * maxlen == TERM_WIDTH) nbcol--;
  if (!nbcol) nbcol = 1;

  pariputc('\n'); i=0;
  for (v=0; list[v]; v++)
  {
    pariputs(list[v]); i++;
    if (i >= nbcol)
    {
      i=0; pariputc('\n');
      if (nbli && j++ > nbli)
      {
	pariputs("---- (type return to continue) ----\n");
	getchar(); j=0;
      }
      continue;
    }
    len = maxlen - strlen(list[v]);
    while (len--) pariputc(' ');
  }
  pariputc('\n');
}

#define LIST_LEN 1023
void
commands(int n)
{
  int hashpos, s = 0, olds = LIST_LEN;
  entree *ep;
  char **list = (char **) xmalloc((olds+1)*sizeof(char *));

  for (hashpos = 0; hashpos < functions_tblsz; hashpos++)
    for (ep = functions_hash[hashpos]; ep; ep = ep->next)
      if ( (n<0 && ep->menu) || ep->menu == n )
      {
        list[s++] = ep->name;
        if (s >= olds)
        {
	  int news = olds + (LIST_LEN + 1)*sizeof(char *);
          xrealloc(list,news,olds);
	  olds = news;
        }
      }
  list[s]=NULL; print_fun_list(list,20); free(list);
}

void
print_user_fun(entree *ep)
{
  GEN *arg = ep->args;
  entree **q = (entree **)(ep->value);

  pariputs(ep->name); pariputc('(');
  while (*arg)
  {
    pariputs((*q)->name); q++;
    pariputc('='); bruteall(*arg,'g',-1,1); arg++;
    if (! *arg) break;
    pariputs(", ");
  }
  pariputs(") = "); pariputs((char*)q);
}

void 
user_fun()
{
  entree *ep;
  int hashpos;

  for (hashpos = 0; hashpos < functions_tblsz; hashpos++)
    for (ep = functions_hash[hashpos]; ep; ep = ep->next)
      if (EpVALENCE(ep) == EpUSER)
      {
	pariputc('{');
	print_user_fun(ep);
	pariputs("}\n\n");
      }
}

void
gentypes(void)
{
 pariputs("\n\
  List of the PARI types :                                              \n\
  ------------------------                                              \n\
  t_INT    : long integers     [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
  t_REAL   : long real numbers [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
  t_INTMOD : integermods       [ code ] [ mod  ] [ integer ]            \n\
  t_FRAC   : irred. rationals  [ code ] [ num. ] [ den. ]               \n\
  t_FRACN  : rational numbers  [ code ] [ num. ] [ den. ]               \n\
  t_COMPLEX: complex numbers   [ code ] [ real ] [ imag ]               \n\
  t_PADIC  : p-adic numbers    [ cod1 ] [ cod2 ] [ p ] [ p^r ] [ int ]  \n\
  t_QUAD   : quadratic numbers [ cod1 ] [ mod  ] [ real ] [ imag ]      \n\
  t_POLMOD : poly mod          [ code ] [ mod  ] [ polynomial ]         \n\
  -------------------------------------------------------------         \n\
  t_POL    : polynomials       [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
  t_SER    : power series      [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
  t_RFRAC  : irred. rat. func. [ code ] [ num. ] [ den. ]               \n\
  t_RFRACN : rational function [ code ] [ num. ] [ den. ]               \n\
  t_QFR    : real qfb          [ code ] [  a  ] [  b  ] [  c  ] [ del ] \n\
  t_QFI    : imaginary qfb     [ code ] [  a  ] [  b  ] [  c  ]         \n\
  t_VEC    : row vector        [ code ] [  x_1  ] ... [  x_k  ]         \n\
  t_COL    : column vector     [ code ] [  x_1  ] ... [  x_k  ]         \n\
  t_MAT    : matrix            [ code ] [ col_1 ] ... [ col_k ]         \n\
\n");
}

void
menu_commands(void)
{
  pariputs("\n\
  1: Standard monadic or dyadic OPERATORS            \n\
  2: CONVERSIONS and similar elementary functions    \n\
  3: TRANSCENDENTAL functions                        \n\
  4: NUMBER THEORETICAL functions                    \n\
  5: Functions related to ELLIPTIC CURVES            \n\
  6: Functions related to general NUMBER FIELDS      \n\
  7: POLYNOMIALS and power series                    \n\
  8: Vectors, matrices, LINEAR ALGEBRA and sets      \n\
  9: SUMS, products, integrals and similar functions \n\
 10: GRAPHIC functions                               \n\
 11: PROGRAMMING under GP                            \n\
\n\
Further help: ?n (1<=n<=11). Also available :      \n\
  ? functionname (short on-line help)              \n\
  ?\\            (keyboard shortcuts)              \n\
  ?.             ( member functions )              \n\
and finally ?? functionname for extended help.\n");
}

void
slash_commands(void)
{
  pariputs("\n\
  #       : enable/disable timer                          \n\
  ##      : print time for last result                    \n\
  \\\\      : comment up to end of line                   \n\
  \\a {n}  : print result in raw format (readable by PARI)\n\
  \\b {n}  : print result in beautified format            \n\
  \\c      : list all commands (same effect as ?*)        \n\
  \\d      : print all defaults                           \n\
  \\e      : enable/disable echo                          \n\
  \\g {n}  : set debugging level                          \n\
  \\k      : kill current GP session but do not exit      \n\
  \\l      : enable/disable logfile                       \n\
  \\m {n}  : print result in prettymatrix format          \n\
  \\p {n}  : change precision                             \n\
  \\q      : quit completely this GP session              \n\
  \\r {f}  : read a file                                  \n\
  \\s {n}  : print stack information                      \n\
  \\t      : print the list of PARI types                 \n\
  \\u      : print the list of user-defined functions     \n\
  \\v      : print current version of GP                  \n\
  \\w {nf} : write to a file                              \n\
  \\x      : print complete inner structure of last result\n\
  \\y      : disable/enable automatic simplification      \n\
  \n\
  {f}=optional filename. {n}=optional integer\n");
}

void
member_commands(void)
{
  pariputs("\n\
  Member functions, followed by relevant objects\n\n\
  a1-a6, b2-b8, c4-c6 : coeff. of the curve.            ell              \n\
  area : area                                           ell              \n\
  bnf  : big number field                                        bnf, bnr\n\
  clgp : classgroup                                              bnf, bnr\n\
  cyc  : cyclic decomposition (SNF)               clgp                   \n\
  diff, codiff : different and codifferent                   nf, bnf, bnr\n\
  disc : discriminant                                   ell, nf, bnf, bnr\n\
  fu   : fundamental units                                       bnf, bnr\n\
  futu : [u,w] where u=unit group, w=torsion                     bnf, bnr\n\
  gen  : generators                               clgp                   \n\
  j    : j-invariant                                    ell              \n\
  nf   : number field                                        nf, bnf, bnr\n\
  no   : number of elements                       clgp                   \n\
  omega, eta : [omega1,omega2] and [eta1, eta2]         ell              \n\
  reg  : regulator                                               bnf, bnr\n\
  roots: roots                                          ell  nf, bnf, bnr\n\
  sign : signature                                           nf, bnf, bnr\n\
  t2   : t2 matrix                                           nf, bnf, bnr\n\
  tu   : torsion units                                           bnf, bnr\n\
  tufu : [w,u] where u=unit group, w=torsion                     bnf, bnr\n\
  zk   : integral basis                                      nf, bnf, bnr\n\
  zkst : structure of (Z_K/m)^* (valid for idealstar also)            bnr\n");
}

#define is_blank_or_null(c) ((!(c) || (c) == ' ' || (c) == '\n'))
#define is_blank(c) ((c) && ((c) == ' ' || (c) == '\n'))

void
print_help(char *s)
{
  long oldlen=0, linelen=0;
  char word[80], oldword[80], *u=word;

  oldword[0]='\0';
  while ((*u++ = *s++))
  {
    if is_blank_or_null(*s)
    {
      while is_blank(*s) s++;
      linelen+=oldlen; if (linelen > 78) { pariputc('\n'); linelen=oldlen; }
      pariputs(oldword); *u++=' '; *u='\0'; oldlen=u-word;
      if (*s) {strcpy(oldword,word);  u=word;}
    }
  }
  *(u-2)='.'; if (linelen+oldlen > 78) pariputc('\n');
  pariputs(word);
}

static void
external_help(char *s)
{
#ifdef GPHELP
  char cmd[80], buf[1024];
  FILE *file;

  sprintf(cmd,"%s '%s'",GPHELP,s);
  file = (FILE *) popen(cmd,"r");
  if (!file)
  {
    sprintf(cmd,"%s %s failed !",GPHELP,s);
    err(talker,cmd);
  }
  while (fgets(buf,1023,file)) 
  {
    buf[1023]=0;
    pariputs(buf);
  }
#else
  err(talker,"Not available in this version");
#endif
}

void
aide(char *s)
{
  long hashpos;
  long n, long_help=0;
  entree *ep;

  if (isdigit(*s))
  {
    n=atoi(s);
    if (n<1 || n > 11)
      err(talker2,"no such section in help: ?",s,s);
    commands(n); return;
  }
  if (*s=='?') {long_help=1; s++;}
  if (*s && !isalpha(*s) && long_help) { external_help(s); return; }
  switch (*s)
  {
    case '*' : commands(-1); return;
    case '\0': menu_commands(); return;
    case '\\': slash_commands(); return;
    case '.' : member_commands(); return;
  }
  hashpos = hashvalue(&s,functions_tblsz,0);
  for(ep = functions_hash[hashpos]; ep; ep = ep->next)
    if(!strcmp(ep->name,s))
    {
      entree *ep1=ep;
      while(EpVALENCE(ep) == EpALIAS) ep = (entree *) ep->value;
      if (ep != ep1) {pariputs(s); pariputs(" is aliased to :\n\n");}
      if (EpVALENCE(ep) == EpUSER) { print_user_fun(ep); return; }
      if (EpVALENCE(ep) == EpVAR) err(talker,"user defined variable");

      if (long_help) { external_help(s); return; }
      if (ep->help) { print_help(ep->help); return; }
      err(talker,"I can't find any help for this. Please report.");
    }
  pariputs("Unknown function\n");
}

typedef struct whatnow_t
{
  char *name;
  char *oldarg;
  char *newarg;
} whatnow_t;

#define STILL NULL
#define OUT (char *) 1

/* generated by PERL script ../util/dico */
const whatnow_t whatnowlist[]={
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"elladd","(e,z1,z2)","(e,z1,z2)"},
{STILL,0,0},
{"matadjoint","(x)","(x)"},
{STILL,0,0},
{"ellak","(e,n)","(e,n)"},
{STILL,0,0},
{"algdep","(x,n,dec)","(x,n,dec)"},
{"nfalgtobasis","(nf,x)","(nf,x)"},
{"ellan","(e,n)","(e,n)"},
{"ellap","(e,n)","(e,n)"},
{"ellap","(e,n)","(e,n,1)"},
{"padicappr","(x,a)","(x,a)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"matcompanion","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{"nfbasis","(x)","(x)"},
{"nfbasis","(x)","(x,2)"},
{"nfbasistoalg","(nf,x)","(nf,x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"ellbil","(e,z1,z2)","(e,z1,z2)"},
{"binomial","(x,y)","(x,y)"},
{STILL,0,0},
{STILL,0,0},
{"contfrac","(x,lmax)","(x,,lmax)"},
{"factor","(x,lim)","(x,lim)"},
{"bnfcertify","(bnf)","(bnf)"},
{"bnfunit","(bnf)","(bnf)"},
{"bnfclassunit","(P)","(P,2)"},
{"bnfclassunit","(P)","(P,1)"},
{"bnfclassunit","(P)","(P)"},
{"quadclassunit","(D,c1,c2,g)","(D,,[c1,c2,g])"},
{"bnfinit","(P)","(P,2)"},
{"bnfinit","(P)","(P,1)"},
{"bnfinit","(P)","(P)"},
{"bnfnarrow","(bnf)","(bnf)"},
{"bnrclass","(bnf,ideal)","(bnf,ideal)"},
{"bnrclass","(bnf,ideal)","(bnf,ideal,1)"},
{"bnrclass","(bnf,ideal)","(bnf,ideal,2)"},
{"quadclassunit","(D)","(D)"},
{"sizebyte","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{"contfrac","(x)","(x)"},
{"contfrac","(b,x)","(x,b)"},
{STILL,0,0},
{"charpoly","(x,y)","(x,y)"},
{"charpoly","(x,y)","(x,y,1)"},
{"charpoly","(x,y)","(x,y,2)"},
{"ellchangecurve","(x,y)","(x,y)"},
{STILL,0,0},
{"ellchangepoint","(x,y)","(x,y)"},
{"qfbclassno","(x)","(x)"},
{"qfbclassno","(x)","(x,1)"},
{"polcoeff","(x,s)","(x,s)"},
{"x*y","(x,y)",""},
{"component","(x,s)","(x,s)"},
{"polcompositum","(pol1,pol2)","(pol1,pol2)"},
{"polcompositum","(pol1,pol2)","(pol1,pol2,1)"},
{"qfbcomprealraw","(x,y)","(x,y)"},
{"vecconcat","(x,y)","(x,y)"},
{"bnfconductor","(a1)","(a1)"},
{"bnfconductorofchar","(bnr,chi)","(bnr,chi)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"serconvol","(x,y)","(x,y)"},
{STILL,0,0},
{"core","(x)","(x,1)"},
{STILL,0,0},
{"coredisc","(x)","(x,1)"},
{STILL,0,0},
{STILL,0,0},
{"truncate","(x)","(x,1)"},
{"polcyclo","(n)","(n)"},
{STILL,0,0},
{"bnfdecodemodule","(nf,fa)","(nf,fa)"},
{"poldegree","(x)","(x)"},
{"denominator","(x)","(x)"},
{"lindep","(x)","(x,-1)"},
{STILL,0,0},
{"matdet","(x)","(x)"},
{"matdet","(x)","(x,1)"},
{"matdetint","(x)","(x)"},
{"matdiagonal","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"poldisc","(x)","(x)"},
{"nfdisc","(x)","(x)"},
{"nfdisc","(x)","(x,2)"},
{"bnrdisc","(bnr,subgroup)","(bnr,subgroup)"},
{"bnrdisc","(bnr)","(bnr,,,2)"},
{"bnrdisclist","(bnf,list)","(bnf,list)"},
{"bnrdisclist","(bnf,arch,bound)","(bnf,bound,arch)"},
{"bnrdisclist","(bnf,bound)","(bnf,bound,,1)"},
{"bnrdisclist","(bnf,bound)","(bnf,bound)"},
{"bnrdisc","(bnr,subgroup)","(bnr,subgroup,,1)"},
{"bnrdisc","(bnr,subgroup)","(bnr,subgroup,,3)"},
{STILL,0,0},
{"divrem","(x,y)","(x,y)"},
{"sumdiv","(n,X,expr)","(n,X,expr)"},
{"mateigen","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"Euler","",""},
{STILL,0,0},
{STILL,0,0},
{"vecextract","(x,y)","(x,y)"},
{"factorial","(x)","(x)"},
{"factorcantor","(x,p)","(x,p)"},
{"factorff","(x,p,a)","(x,p,a)"},
{"factormod","(x,p)","(x,p)"},
{STILL,0,0},
{"nfbasis","(x,p)","(x,,p)"},
{"nfdisc","(x,p)","(x,,p)"},
{"polred","(x,p)","(x,,p)"},
{"polred","(x,p)","(x,2,p)"},
{STILL,0,0},
{STILL,0,0},
{"factorpadic","(x,p,r)","(x,p,r,1)"},
{"factor","(x,l,hint)","(x)"},
{"factor","(x,l,hint)","(x)"},
{"fibonacci","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"ffinit","(p,n)","(p,n)"},
{STILL,0,0},
{"polgalois","(x)","(x)"},
{"nfgaloisapply","(nf,aut,x)","(nf,aut,x)"},
{"nfgaloisconj","(nf)","(nf)"},
{"nfgaloisconj","(nf)","(nf,2)"},
{"nfgaloisconj","","(nf,1)"},
{"gammah","(x)","(x)"},
{STILL,0,0},
{"matsolve","(a,b)","(a,b)"},
{"matsolvemod","(M,D,Y)","(M,D,Y)"},
{"matsolvemod","(M,D,Y)","(M,D,Y,1)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"ellglobalred","(x,y)","(x,y)"},
{STILL,0,0},
{"qfbhclassno","(x)","(x)"},
{"ellheight","(e,x)","(e,x)"},
{"ellheight","(e,x)","(e,x,1)"},
{"mathnf","(x)","(x,1)"},
{"mathnf","(x)","(x,1)"},
{"mathnf","(x)","(x,2)"},
{"hnfmod","(x,d)","(x,d)"},
{"hnfmodid","(x,d)","(x,d)"},
{"mathnf","(x)","(x,3)"},
{"mathess","(x)","(x)"},
{"hilbert","(x,y)","(x,y)"},
{"mathilbert","(n)","(n)"},
{"hilbert","(x,y,p)","(x,y,p)"},
{"vector","(n,X,expr)","(n,X,expr)"},
{STILL,0,0},
{"I","",""},
{STILL,0,0},
{"idealaddone","(nf,list)","(nf,list)"},
{STILL,0,0},
{STILL,0,0},
{"idealappr","(nf,x)","(nf,x,1)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"idealdiv","(nf,x,y)","(nf,x,y,1)"},
{STILL,0,0},
{"idealhnf","(nf,x)","(nf,x)"},
{"idealhnf","(nf,x)","(nf,x)"},
{STILL,0,0},
{STILL,0,0},
{"idealinv","(nf,x)","(nf,x,1)"},
{STILL,0,0},
{STILL,0,0},
{"ideallistarch","(nf,list,arch)","(nf,list,arch,1)"},
{"ideallist","(nf,list)","(nf,list,2)"},
{"ideallistarch","","(nf,list,arch,2)"},
{"ideallistarch","","(nf,list,arch,3)"},
{"ideallist","","(nf,list,3)"},
{"ideallist","(nf,bound)","(nf,bound)"},
{"ideallist","(nf,bound)","(nf,bound,1)"},
{"idealred","(nf,x,vdir)","(nf,x,vdir)"},
{STILL,0,0},
{"idealmul","(nf,x,y)","(nf,x,y,1)"},
{STILL,0,0},
{STILL,0,0},
{"idealpow","(nf,x,y)","(nf,x,y,1)"},
{STILL,0,0},
{"idealtwoelt","(nf,x,a)","(nf,x,a)"},
{STILL,0,0},
{"matid","(n)","(n)"},
{STILL,0,0},
{STILL,0,0},
{"matimage","(x)","(x)"},
{"matimage","(x)","(x,1)"},
{"matimagecompl","(x)","(x)"},
{STILL,0,0},
{OUT,0,0},
{OUT,0,0},
{OUT,0,0},
{"incgam","(s,x,y)","(s,x,y)"},
{"matindexrank","(x)","(x)"},
{"vecindexsort","(x)","(x)"},
{"nfinit","(pol)","(pol)"},
{"nfinit","(x)","(x,2)"},
{"nfinit","(x)","(x,3)"},
{"ellinit","(x)","(x)"},
{"zetakinit","(x)","(x)"},
{"intformal","(x,y)","(x,y)"},
{"matintersect","(x,y)","(x,y)"},
{"intnum","(x=a,b,s)","(x=a,b,s,1)"},
{"intnum","(x=a,b,s)","(x=a,b,s,2)"},
{STILL,0,0},
{"intnum","(x=a,b,s)","(x=a,b,s,3)"},
{"matinverseimage","(x,y)","(x,y)"},
{"matisdiagonal","(x)","(x)"},
{"isfundamental","(x)","(x)"},
{"nfisideal","(nf,x)","(nf,x)"},
{"nfisincl","(x,y)","(x,y)"},
{"nfisincl","(nf1,nf2)","(nf1,nf2,1)"},
{"polisirreducible","(x)","(x)"},
{"nfisisom","(x,y)","(x,y)"},
{"nfisisom","(nf1,nf2)","(nf1,nf2,1)"},
{"ellisoncurve","(e,x)","(e,x)"},
{STILL,0,0},
{"bnfisprincipal","(bnf,x)","(bnf,x,0)"},
{"bnfisprincipal","(bnf,x)","(bnf,x,2)"},
{"bnfisprincipal","(bnf,x)","(bnf,x)"},
{"bnfisprincipal","(bnf,x)","(bnf,x,3)"},
{"bnrisprincipal","(bnf,x)","(bnf,x)"},
{STILL,0,0},
{"ispseudoprime","(x)","(x)"},
{"sqrtint","(x)","(x)"},
{"setisset","(x)","(x)"},
{"issquarefree","(x)","(x)"},
{STILL,0,0},
{"bnfisunit","(bnf,x)","(bnf,x)"},
{"qfjacobi","(x)","(x)"},
{"besseljh","(n,x)","(n,x)"},
{"ellj","(x)","(x)"},
{"polkaramul","(x,y,k)","(x,y,k)"},
{"besselk","(nu,x)","(nu,x)"},
{"besselk","(nu,x)","(nu,x,1)"},
{"matker","(x)","(x)"},
{"matker","(x)","(x,1)"},
{"matkerint","(x)","(x)"},
{"matkerint","(x)","(x,1)"},
{"matkerint","(x)","(x,2)"},
{"kronecker","(x,y)","(x,y)"},
{STILL,0,0},
{"zetak","(nfz,s)","(nfz,s,1)"},
{"serlaplace","(x)","(x)"},
{STILL,0,0},
{"pollegendre","(n)","(n)"},
{STILL,0,0},
{STILL,0,0},
{"veclexsort","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{"lindep","(x)","(x,1)"},
{"qflll","(x)","(x)"},
{"qflll","(x)","(x,7)"},
{"qflll","(x)","(x,8)"},
{"qflllgram","(x)","(x)"},
{"qflllgram","(x)","(x,7)"},
{"qflllgram","(x)","(x,8)"},
{"qflllgram","(x)","(x,1)"},
{"qflllgram","(x)","(x,4)"},
{"qflllgram","(x)","(x,5)"},
{"qflll","(x)","(x,1)"},
{"qflll","(x)","(x,2)"},
{"qflll","(x)","(x,4)"},
{"qflll","(x)","(x,5)"},
{"qflll","(x)","(x,3)"},
{"log","(x)","(x)"},
{STILL,0,0},
{"elllocalred","(e)","(e)"},
{STILL,0,0},
{"log","(x)","(x,1)"},
{"elllseries","(e,s,N,A)","(e,s,N,A)"},
{"bnfmake","(sbnf)","(sbnf)"},
{"Mat","(x)","(x)"},
{"vecextract","(x,y,z)","(x,y)"},
{"ellheightmatrix","(e,x)","(e,x)"},
{STILL,0,0},
{STILL,0,0},
{"matrixqz","(x,p)","(x,-1)"},
{"matrixqz","(x,p)","(x,-2)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"idealmin","(nf,ix,vdir)","(nf,ix,vdir)"},
{"qfminim","(x,bound,maxnum)","(x,bound,maxnum)"},
{"qfminim","(x,bound)","(x,bound,,1)"},
{"Mod","(x,y)","(x,y)"},
{"Mod","(x,y,p)","(x,y,1)"},
{STILL,0,0},
{"gcd","(x,y)","(x,y,1)"},
{"moebius","(n)","(n)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"nfeltdiv","(nf,a,b)","(nf,a,b)"},
{"nfeltdiveuc","(nf,a,b)","(nf,a,b)"},
{"nfeltdivrem","(nf,a,b)","(nf,a,b)"},
{"nfhnf","(nf,x)","(nf,x)"},
{"nfhnfmod","(nf,x,detx)","(nf,x,detx)"},
{"nfeltmod","(nf,a,b)","(nf,a,b)"},
{"nfeltmul","(nf,a,b)","(nf,a,b)"},
{"nfeltpow","(nf,a,k)","(nf,a,k)"},
{"nfeltreduce","(nf,a,id)","(nf,a,id)"},
{"nfsnf","(nf,x)","(nf,x)"},
{"nfeltval","(nf,a,pr)","(nf,a,pr)"},
{STILL,0,0},
{STILL,0,0},
{"qfbnucomp","(x,y,l)","(x,y,l)"},
{STILL,0,0},
{"numerator","(x)","(x)"},
{"qfbnupow","(x,n)","(x,n)"},
{"O","(x)","(x)"},
{STILL,0,0},
{"ellordinate","(e,x)","(e,x)"},
{"znorder","(x)","(x)"},
{"ellorder","(e,x)","(e,x)"},
{"polredord","(x)","(x)"},
{STILL,0,0},
{"matpascal","(n)","(n)"},
{"qfperfection","(a)","(a)"},
{"numtoperm","(n,k)","(n,k)"},
{"permtonum","(vect)","(vect)"},
{"qfbprimeform","(x,p)","(x,p)"},
{"eulerphi","(x)","(x)"},
{"Pi","",""},
{"contfracpnqn","(x)","(x)"},
{"ellztopoint","(e,z)","(e,z)"},
{"polinterpolate","(xa,ya,x)","(xa,ya,p)"},
{STILL,0,0},
{"polred","(x)","(x,2)"},
{STILL,0,0},
{"polredabs","(x)","(x,1)"},
{"polredabs","(x)","(x,4)"},
{"polredabs","(x)","(x,8)"},
{"polredabs","(x)","(x,2)"},
{STILL,0,0},
{"variable","(x)","(x)"},
{"Pol","(x,v)","(x,v)"},
{STILL,0,0},
{"polylog","(m,x)","(m,x,1)"},
{"polylog","(m,x)","(m,x,2)"},
{"polylog","(m,x)","(m,x,3)"},
{"Polrev","(x,v)","(x,v)"},
{"polzagier","(n,m)","(n,m)"},
{"ellpow","(e,x,n)","(e,x,n)"},
{"qfbpowrealraw","(x,n)","(x,n)"},
{"precision","(x,n)","(x,n)"},
{STILL,0,0},
{STILL,0,0},
{"idealprimedec","(nf,p)","(nf,p)"},
{STILL,0,0},
{"znprimroot","(n)","(n)"},
{"idealprincipal","(nf,x)","(nf,x)"},
{"ideleprincipal","(nf,x)","(nf,x)"},
{"prod","(x,X=a,b,expr)","(X=a,b,expr,x)"},
{STILL,0,0},
{STILL,0,0},
{"prodinf","(X=a,expr)","(X=a,expr,1)"},
{STILL,0,0},
{"Qfb","(a,b,c)","(a,b,c)"},
{"Qfb","(a,b,c,d)","(a,b,c,d)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"matrank","(x)","(x)"},
{"bnrclassno","(bnf,x)","(bnf,x)"},
{"bnrclassnolist","(bnf,liste)","(bnf,liste)"},
{STILL,0,0},
{"polrecip","(x)","(x)"},
{"qfbred","(x)","(x)"},
{"qfbred","(x)","(x)"},
{"qfbred","(x)","(x,2)"},
{"poldiscreduced","(f)","(f)"},
{"quadregulator","(x)","(x)"},
{STILL,0,0},
{"polresultant","(x,y)","(x,y)"},
{"polresultant","(x,y)","(x,y,1)"},
{"serreverse","(x)","(x)"},
{"qfbred","(x)","(x,1)"},
{"qfbred","(x)","(x,3)"},
{"round","(x)","(x,1)"},
{STILL,0,0},
{"rnfdisc","(nf,pol)","(nf,pol)"},
{STILL,0,0},
{"rnfequation","(nf,pol)","(nf,pol,1)"},
{"rnfhnfbasis","(bnf,order)","(bnf,order)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"polrootsmod","(x,p)","(x,p)"},
{"rootmod2","(x,p)","(x,p,1)"},
{"polrootspadic","(x,p,r)","(x,p,r)"},
{"polroots","(x)","(x)"},
{"nfrootsof1","(nf)","(nf)"},
{"polroots","(x)","(x,1)"},
{STILL,0,0},
{STILL,0,0},
{"Ser","(x,v)","(x,v)"},
{"Set","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"sigma","(x,k)","(x,k)"},
{STILL,0,0},
{"qfsign","(x)","(x)"},
{"bnfsignunit","(bnf)","(bnf)"},
{"factormod","(x,p)","(x,p,1)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"sizedigit","(x)","(x)"},
{"nfbasis","(x)","(x,1)"},
{"bnfinit","(x)","(x,3)"},
{"nfdiscf","","(x,1)"},
{"factor","(x)","(x,0)"},
{"initell","(x)","(x,1)"},
{"polred","(x)","(x,1)"},
{"polred","(x)","(x,3)"},
{"matsnf","(x)","(x)"},
{"matsnf","(x)","(x,1)"},
{"matsnf","(x)","(x,4)"},
{"matsnf","(x)","(x,2)"},
{STILL,0,0},
{"vecsort","(x)","(x)"},
{STILL,0,0},
{"qfgaussred","(x)","(x)"},
{STILL,0,0},
{"gcd","(x,y)","(x,y,2)"},
{"polsturm","(x)","(x)"},
{"polsturmpart","(x,a,b)","(x,a,b)"},
{"polsubcyclo","(p,d)","(p,d)"},
{"ellsub","(e,a,b)","(e,a,b)"},
{STILL,0,0},
{"sum","(x,X=a,b,expr)","(X=a,b,expr,x)"},
{STILL,0,0},
{"sumalt","(X=a,expr)","(X=a,expr,1)"},
{STILL,0,0},
{STILL,0,0},
{"sumpos","(X=a,expr)","(X=a,expr,1)"},
{"matsupplement","(x)","(x)"},
{"polsylvestermatrix","(x,y)","(x,y)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"elltaniyama","(e)","(e)"},
{STILL,0,0},
{"poltchebi","(n)","(n)"},
{"teichmuller","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{OUT,0,0},
{OUT,0,0},
{"elltors","(e)","(e)"},
{STILL,0,0},
{"mattranspose","(x)","(x)"},
{"truncate","(x)","(x)"},
{"poltschirnhaus","(x)","(x)"},
{OUT,0,0},
{"quadunit","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{"Vec","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{STILL,0,0},
{"vectorv","(n,X,expr)","(n,X,expr)"},
{"ellwp","(e)","(e)"},
{"weber","(x)","(x)"},
{"weber","(x)","(x,2)"},
{STILL,0,0},
{"ellpointtoz","(e,P)","(e,P)"},
{STILL,0,0},
{STILL,0,0},
{"ideallog","(nf,x,bid)","(nf,x,bid)"},
{"idealstar","(nf,I)","(nf,I)"},
{"idealstar","(nf,id)","(nf,id,1)"},
{"idealstar","(nf,id)","(nf,id,2)"},
{STILL,0,0},

{STILL,0,0},
{"plotbox","(x,a)","(x,a)"},
{"plotcolor","(w,c)","(w,c)"},
{"plotcursor","(w)","(w)"},
{STILL,0,0},
{"plotdraw","(list)","(list)"},
{"plotinit","(w,x,y)","(w,x,y)"},
{STILL,0,0},
{"plotkill","(w)","(w)"},
{"plotlines","(w,x2,y2)","(w,x2,y2)"},
{"plotlines","(w,x2,y2)","(w,x2,y2)"},
{"plotmove","(w,x,y)","(w,x,y)"},
{STILL,0,0},
{STILL,0,0},
{"ploth","(X=a,b,expr)","(X=a,b,expr,1)"},
{"ploth","(X=a,b,expr)","(X=a,b,expr)"},
{STILL,0,0},
{"plotpoints","(w,x,y)","(w,x,y)"},
{"plotpoints","(w,x,y)","(w,x,y)"},
{"psdraw","(list)","(list)"},
{"psploth","(X=a,b,expr)","(X=a,b,expr)"},
{"psploth","(X=a,b,expr)","(X=a,b,expr,1)"},
{"psplothraw","(listx,listy)","(listx,listy)"},
{"printp","(x)","(x)"},
{"printp1","(x)","(x)"},
{STILL,0,0},
{STILL,0,0},
{"input","(x)","(x)"},
{"plotrbox","(w,dx,dy)","(w,dx,dy)"},
{"plotrline","(w,dx,dy)","(w,dx,dy)"},
{"plotrlines","(w,dx,dy)","(w,dx,dy,1)"},
{"plotrmove","(w,dx,dy)","(w,dx,dy)"},
{"plotrpoint","(w,dx,dy)","(w,dx,dy)"},
{"plotrpoints","(w,dx,dy)","(w,dx,dy)"},
{"plotscale","(w,x1,x2,y1,y2)","(w,x1,x2,y1,y2)"},
{"default","(n)","(realprecision,n)"},
{"default","(n)","(seriesprecision,n)"},
{"type","(x,t)","(x,t)"},
{"plotstring","(w,x)","(w,x)"},
{"printtex","(x)","(x)"},
{"type","(x,t)","(x,t)"},
{STILL,0,0},
};

/* 
 * If silent != 0: silently return 1 if function changed, 0 otherwise.
 * Else be verbose.
 */
int
whatnow(char *s, int silent)
{
  long n=0, len=strlen(s);
  char *def, str[128];
  whatnow_t wp;
  entree *ep;

  do {def = (oldfonctions[n++]).name;}
  while (def && (strncmp(def,s,len)));
  if (!def)
  {
    int m=0;
    do {def = (functions_oldgp[m++]).name;}
    while (def && (strncmp(def,s,len)));
    n += m - 1;
  }

  if (def && strlen(def) == len)
  {
    wp=whatnowlist[n-1]; def=wp.name;
    switch((int) def)
    {
      case (int) STILL:
	if (silent) return 0;
	err(talker,"This function did not change !");
      case (int) OUT: 
	if (silent) return 1;
	err(talker,"This function was suppressed !");
    }
    if (silent) return 1;
    if (!strcmp(def,"x*y"))
    {
      sprintf(str,"  %s is now called *.\n\n",s);
      pariputs(str);
      sprintf(str,"    %s%s ===> %s%s\n\n",s,wp.oldarg,wp.name,wp.newarg);
      pariputs(str); return 1;
    }
    n = hashvalue(&def,functions_tblsz,1);
    ep = findentry(n,wp.name,def,functions_hash);
    if (!ep) err(talker,"Bug in whatnow. Please report !");
    if (strcmp(s,ep->name))
      sprintf(str,"  %s is now called %s.\n\n",s,ep->name);
    else
      sprintf(str,"  The syntax changed:\n\n");
    pariputs(str);
    sprintf(str,"    %s%s ===> %s%s\n\n",s,wp.oldarg,wp.name,wp.newarg);
    pariputs(str); print_help(ep->help); pariputs("\n\n");
    return 1;
  }
  if (silent) return 0;
  err(talker,"As far as I can recall, this one never existed !");
  return 0; /* not reached */
}

void
read0(char *s)
{
  switchin_byname(s); gp_main_loop(gpsilent ? SILENT : REGULAR);
}

void
extern0(char *cmd)
{
  FILE *file = try_pipe(cmd);
 
  switchin(file); gp_main_loop(gpsilent ? SILENT : REGULAR);
}

void
default0(char *a, char *b)
{
  setdefault(a,b, gpsilent? SILENT: ACKNOWLEDGE);
}

GEN
allocatemem0(unsigned long newsize)
{
  allocatemoremem(newsize);
  parisize=newsize;
  longjmp(environnement,errpile);
  return NULL; /* not reached */
}

/* s is a string object, i.e a mix of real strings, and GENs (the adresses),
 * whose first sizeof(long) chars contain the total number of objects (i.e of
 * connected components, separated by some NULL chars). All GENs are preceded
 * by \0\0 (to distinguish from the single ending \0 of strings). The first
 * object if it's a GEN, takes only one \0)
 */
void
print0(char *s, long flag)
{ 
  int old=prettyp, newline;
  long nb; /* number of objects */

  if (flag > 3) { flag -= 4; newline=0; } else newline=1;
  prettyp=flag;

  memcpy(&nb,s,sizeof(long)); s += sizeof(long);
  for ( ; nb>0; nb--)
  {
    if (*s)
    {
      pariputs(s); while (*s++);
    }
    else
    {
      GEN res;

      s++; /* skip \0 */
      memcpy(&res,s,sizeof(GEN)); s += sizeof(GEN);
      gp_output(res);
      if (!*s) s++; /* next one is a GEN */
    }
  }
  if (newline) pariputc('\n');
  fflush(outfile); if (logfile) fflush(logfile);
  prettyp=old;
}

GEN
input0()
{
  char *buffer = (char *) xmalloc(paribuffsize);
  GEN res;

  while (!fgets(buffer, paribuffsize, infile)) switchin(NULL);
  if (pariecho) pariputs(buffer);
  else if (logfile) fputs(buffer, logfile);
  res=lisseq(buffer); free(buffer);

  return res;
}

GEN
type0(GEN x, long flag)
{
  if (flag)
  { 
    GEN y=gcopy(x);
    
    settyp(y,flag); return y;
  }
  return stoi(typ(x));
}

GEN
type00(GEN x, char *st)
{
  char msg[80];

  if (! *st) return type0(x,0);
  if (isdigit(*st))
  {
    char *s = st;
    while (*s && isdigit(*s)) s++;
    if (! *s) return type0(x,atol(st));
  }

  if (!strncmp(st,"t_",2)) st += 2; /* skip initial part */

  switch(strlen(st))
  {
    case 3:
      if (!strcmp(st,"INT")) return type0(x,t_INT);
      if (!strcmp(st,"POL")) return type0(x,t_POL);
      if (!strcmp(st,"SER")) return type0(x,t_SER);
      if (!strcmp(st,"QFR")) return type0(x,t_QFR);
      if (!strcmp(st,"QFI")) return type0(x,t_QFI);
      if (!strcmp(st,"VEC")) return type0(x,t_VEC);
      if (!strcmp(st,"COL")) return type0(x,t_COL);
      if (!strcmp(st,"MAT")) return type0(x,t_MAT);
      break;

    case 4:
      if (!strcmp(st,"REAL")) return type0(x,t_REAL);
      if (!strcmp(st,"FRAC")) return type0(x,t_FRAC);
      if (!strcmp(st,"QUAD")) return type0(x,t_QUAD);
      break;

    case 5:
      if (!strcmp(st,"FRACN")) return type0(x,t_FRACN);
      if (!strcmp(st,"PADIC")) return type0(x,t_PADIC);
      if (!strcmp(st,"RFRAC")) return type0(x,t_RFRAC);
      break;

    case 6:
      if (!strcmp(st,"INTMOD")) return type0(x,t_INTMOD);
      if (!strcmp(st,"POLMOD")) return type0(x,t_POLMOD);
      if (!strcmp(st,"RFRACN")) return type0(x,t_RFRACN);
      break;

    case 7:
      if (!strcmp(st,"COMPLEX")) return type0(x,t_COMPLEX);
      break;
  }
  sprintf(msg,"Unknown type : t_%s",st);
  err(talker,msg);
  return NULL; /* not reached */
}

void
write0(long d,char *s)
{
  if (d>tglobal)
  {
    char str[50];
    sprintf(str,"write(%d,...)",(int)d);
    err(referer2,str+6,str);
  }
  if (!*s) s=current_logfile;
  switchout(s);
  bruteall(g[d?d:tglobal],fmt.format,-1,0);
  pariputc('\n'); switchout(NULL);
}

void
system0(char *s)
{
#ifdef UNIX
  system(s);
#else
  err(talker,"Not available on this architecture");
#endif
}

void
error0()
{
  err_recover(talker);
}

void
addhelp(entree *ep, char *s)
{
  if (ep->help && !(ep->valence & EpSTATIC))
  {
    free(ep->help);
    ep->valence &= ~EpSTATIC;
  }
  ep->help = (char*) xmalloc(strlen(s) + 1);
  strcpy(ep->help, s);
}

long
setprecr(long n)
{
  long m = fmt.nb;

  if(n>0) {fmt.nb = n; prec = (long)(n*K1 + 3);}
  return m;
}
